SFPhoton modular web application
================================

Available at: [sferi.hu](https://sferi.hu/).

SFPhoton modular web application is a small web framework that implements a website, relying on modules for easy extension. This is the core of my personal website with a diverse portfolio of services - and my first larger project in Flask.

Looking back, I would have been better off using Flask Blueprints instead of modules, but hey, this approach works, too. I don't think many people should rely on mysolution but I decided to publish the source as only time can tell if it will help anyone.

Usage
-----

### Dependencies

- python
- mongodb
- some python libs (install them via `pip install -r requirements.txt`)

### Run

Initialize the database with the `db_manadement/init_db.py` script.

In debug mode: `./run-debug.sh`

In production mode: please refer to the [Flask deployment guide](https://flask.palletsprojects.com/en/3.0.x/deploying/) for a decent deploy. After that, you should run `./run.sh`.

Modules
-------

This web application has minimal code as its core, and most of its functionality is implemented as modules. A module can provide endpoints (e.g. /my-module-functionality) and menu items for users added to it. Authentication and authorization is automatic, but endpoints may implement it, too. The modules are stored in the `plugins` folder in separate subfolders. SFPhoton modular web application will add any module found in the `plugins` folder on start.

### Existing modules

For some open-source modules that I maintain, look into my projects here. Some recommended ones both for using and for reference:

- **[Profile](https://git.sch.bme.hu/schulczf/sfphoton-web-plugin-profile)**: a module to let users change their passwords. Useful if you have local users.
- **[Minecraft](https://git.sch.bme.hu/schulczf/sfphoton-web-plugin-minecraft)**: a feature-rich web ecosystem for a Minecraft server, with live map viewing, player authorization and an admin interface.
- **[DynDNS](https://git.sch.bme.hu/schulczf/sfphoton-web-plugin-dyndns)**: frontend for [my DynDNS server](https://git.sch.bme.hu/schulczf/dyndns-server).
- **[Dummy Cat](https://git.sch.bme.hu/schulczf/sfphoton-web-plugin-dummy_cat)**: a dead simple module that showed an AI-generated cat until thiscatdoesnotexist.com went offline. Still, an easy example module.

### Installing a module

Just copy its folder into `plugins` and restart SFPhoton modular web application. If you want to use an open-source module from the previous section, just navigate to `plugins` and clone the appropriate repository.

### Creating your module

#### Layout of a plugin

Your plugin should be a subfolder in the `plugins` folder. At least, your plugin folder should have a Python file (e. g. `myplugin.py`) to implement the logic and endpoints for your module.

Once your plugin grows, it is a good idea to add further subfolder for templates and static files. After all, your layout will probably look like this:

```
plugins/
 |\_myplugin.py
 |\_templates/
 |   \_myview.html
  \_static
    |\_image1.png
     \_style.css
```

#### How to code a plugin

Your plugin should do two things:

1. Provide endpoints for HTTP requests
1. Register those endpoints in SFPhoton modular web application

_Creating endpoints_

Every endpoint should be a plain old Python function that only takes `**kwargs` as a parameter. `**kwargs` is a dictionary which will hold 3 things:

- `kwargs['session']` is the Flask `session` object
- `kwargs['request']` is the Flask `request` object
- `kwargs['rqtools']` is an object that has several useful functions
    - `url_for()` creates a URL from an endpoint description, just as in Flask. To generate URLs for plugin endpoints, use `url_for('service', servicename='endpoint_name')` where `endpoint_name` is the name for the endpoint you register (like `/endpoint_name`), the others should remain constant.
    - `redirect()` is the same as Flask's `redirect()` function.
    - `get_xxx()` functions are to show HTTP error pages to users, like `get_404(None)` for a simple 404 Not found screen. You can print custom error messages like `get_404(None, "Custom error message.")`
    - `render_template()` can be used to render any jinja2 template like with Flask's `render_template()` function, but template locations are not limited to the framework's template folder. Instead, you have to supply the path to your template file like `render_template('/path/to/template')`. Keyword arguments are optional like with Flask. **Warning**: you should sanitize your template file locations to prevent path traversal attacks.

Endpoints are expected to return the same thing as Flask endpoints would do. Thus, an example endpoint which renders a template based on a URL parameter would look like this:

```python
import os

def kitten_picture(**kwargs):
    request = kwargs['request']
    rqtools = kwargs['rqtools']

    # If no kitten name is provided as URL parameter, respond with HTTP 400
    if 'kitten-name' not in request.args.keys():
        return rqtools.get_400(None, "No kitten name provided!")

    # Render kitten template from the plugin folder's templates subfolder
    # Provide kitten name as an argument for the template
    template_path = os.path.join(os.path.dirname(__file__), 'templates', 'kitten.html')
    return rqtools.render_template(template_path, kitten=request.args['kitten-name'])
```

_Registering endpoints_

Endpoints should be registered to handle requests or to show up in the menu. For this, every endpoint has to provide a `register()` function, which will be called on startup.

This function should have exactly one parameter, in which it will be provided a pointer to the `add_endpoint()` function it can call to configure endpoints.

Expected behaviour of the `register()` function is best shown on an example:

```python
def register(add_endpoint):

    # Do any initialization for you plugin here
    ...

    # Register endpoints like this:

    # Public endpoint, available for GET requests at /kittens
    # For this to work, you should have a function called kitten_picture() that implements the enpoint (see above)
    add_endpoint(endpoint_id='kittenimage', handler=kitten_picture, permission_name=None)

    # Public endpoint, available for PUT requests at /newkitten
    add_endpoint(endpoint_id='newkitten', handler=new_kitten, permission_name=None, method='PUT')

    # Public endpoint but with a menu item
    add_endpoint(endpoint_id='kittenworld', handler=kittens_mainpage, permission_name=None, menutext='Kitten world')

    # Endpoint only for logged-in users that have the 'kittenadmin' permission
    # This also registers a menu item that is visible only for authorized users
    add_endpoint(endpoint_id='kittenadmin', handler=kittens_adminpage, permission_name='kittenadmin', menutext='Kitten admin')

    # Another public endpoint, but for kitten admins this menu item is grouped together with 'Kitten admin'
    add_endpoint(endpoint_id='kittenchaos', handler=kittens_chaos_game, permission_name=None, menutext='Kitten chaos', menu_permission_hint='kittenadmin')
```

User management
---------------

Look into the `db_management` folder for a bunch of scripts.