from flask import (Flask, render_template, render_template_string, send_from_directory, redirect,
                   session, request, flash, url_for, send_file)
from werkzeug.security import check_password_hash
from os.path import dirname, basename, isfile, join
import os
import glob
import random
import sys
import importlib.util
import requests
from authlib.integrations.flask_client import OAuth

import services
import db
import config

# Create Flask app
###############################################################

app = Flask("sfphotonweb")

config.init()
app.secret_key = config.get('FLASK_SECRET')

@app.before_request
def make_session_permanent():
    session.permanent=True


# Set up endpoint database
###############################################################

public_service_endpoints = dict()
public_service_menu = []
authorized_service_endpoints = dict()
authorized_service_menu = []

def plugin_add_endpoint(endpoint_id: str, handler, permission_name: str, method='GET', menutext=None, menu_permission_hint=None):
    """Add a dynamic endpoint."""
    """  id: used in the url (like /something)"""
    """  handler: pointer to handler function"""
    """  servicename: service identifier for authorization. If None, it is a public endpoint"""
    """  method: HTTP method this endpoint can handle. If you plan to handle multiple methods with one endpoint, call this function multiple times. (Make sure to only add a menu once.)"""
    """  menutext: Text to appear in the menu. If None, no menu item is added."""
    """  menu_permission_hint: used only at public endpoints with menu item. If you want to place your menu item next to another (but authorized) menu item, specify the same permission name here. Public endpoints with menu_permission_hint=None will be automatically placed together for each plugin at the end of the menu strip"""
    
    if permission_name == None:
        # Public endpoint
        # add endpoint
        if not method in public_service_endpoints:
            public_service_endpoints[method] = dict()
        public_service_endpoints[method][endpoint_id] = handler
        # add to menu
        if not menutext == None:
            public_service_menu.append((menutext, endpoint_id))
            authorized_service_menu.append((menutext, endpoint_id, menu_permission_hint))
    else:
        # Authorized endpoint
        # add endpoint
        if not method in authorized_service_endpoints:
            authorized_service_endpoints[method] = dict()
        authorized_service_endpoints[method][endpoint_id] = (permission_name, handler)
        # add to menu
        if not menutext == None:
            authorized_service_menu.append((menutext, endpoint_id, permission_name))


# Import and register plugins
###############################################################

for plugin_dir in os.listdir(join(dirname(__file__)+'/plugins')):
    plugin_dir_path = join(dirname(__file__)+'/plugins/'+plugin_dir)
    if os.path.isdir(plugin_dir_path):
        for plugin_file in os.listdir(plugin_dir_path):
            plugin_file_path = join(plugin_dir_path, plugin_file)
            if os.path.isfile(plugin_file_path) and plugin_file_path.endswith('.py'):
                plugin = "plugins." + plugin_dir + '.' + plugin_file[:-3]
                print("Importing plugin " + plugin)
                spec = importlib.util.spec_from_file_location(plugin, plugin_file_path)
                module = importlib.util.module_from_spec(spec)
                sys.modules[plugin] = module
                spec.loader.exec_module(module)

                # calling register()
                if hasattr(module, "register"):
                    module.register(plugin_add_endpoint)
                else:
                    print('  ' + plugin + ' has no register() function.')



# Authentication stuff
###############################################################

def find_real_username(username):
    x = db.users.find_one(filter={'username': username})
    if 'redirectname' in x:
        return find_real_username(x['redirectname'])
    return username


oauth = OAuth(app)
oauth.register(
    name='authsch',
    client_id=config.get('AUTHSCH_ID'),
    client_secret=config.get('AUTHSCH_SECRET'),
    access_token_url='https://auth.sch.bme.hu/oauth2/token',
    access_token_params=None,
    authorize_url='https://auth.sch.bme.hu/site/login',
    authorize_params=None,
    client_kwargs={
        'scope': 'linkedAccounts', },
)


# Routing
###############################################################

@app.route('/')
def index():
    return render_template('index.html')


@app.route('/neptun')
def neptun():
    return render_template('neptun.html')


@app.route('/login', methods=('GET', 'POST'))
def login():
    if request.method == 'POST':
        username = request.form['username']
        password = request.form['password']
        x = db.users.find_one(filter={'username': username})
        if x is None:
            flash('No such user.')
            return render_template('login.html'), 401
        if not check_password_hash(x['pwhash'], password):
            flash('Password is incorrect')
            return render_template('login.html'), 401
        session.clear()
        app.logger.info("%s logged in", x['username'])
        session['username'] = find_real_username(x['username'])
        if 'next' in request.args.keys():
            return redirect(request.args['next'])
        return redirect(url_for('index'))
    return render_template('login.html')


@app.route('/authsch_login')
def authsch_login():
    redirect_uri = url_for('authsch_authorize', _external=True)
    return oauth.authsch.authorize_redirect(redirect_uri)


@app.route('/authsch_authorize')
def authsch_authorize():
    token = oauth.authsch.authorize_access_token()
    resp = requests.get(
        'https://auth.sch.bme.hu/api/profile/?access_token='+token['access_token'])
    resp.raise_for_status()
    print(resp.json())
    profile = resp.json()
    uname = ""
    if 'schacc' in profile['linkedAccounts']:
        uname = 'schacc:' + profile['linkedAccounts']['schacc']
    elif 'bme' in profile['linkedAccounts']:
        uname = 'bme:' + profile['linkedAccounts']['bme']
    elif 'vir' in profile['linkedAccounts']:
        uname = 'vir:' + profile['linkedAccounts']['vir']
    else:
        uname = 'authsch:' + profile['internal_id']
    
    # add services
    x = db.users.find_one(filter={'username': uname})
    if x is None:
        userobj = {"username": uname,
                   "pwhash": 'None', "services": ['dyndns', 'minecraft']}
        db.users.insert_one(userobj)

    # set up session
    session.clear()
    app.logger.info("%s logged in", uname)
    session['username'] = find_real_username(uname)

    return redirect(url_for('index'))


@app.route('/logout')
def logout():
    if 'username' in session:
        app.logger.info("%s logged out", session['username'])
    session.clear()
    return redirect(url_for('index'))


@app.route('/robots.txt')
def robotstxt():
    return send_file(join(dirname(__file__))+'/static/robots.txt')


@app.errorhandler(404)
def get_404(error, **kwargs):
    messages = [
        "Maybe someone borrowed it?",
        "Maybe it's under the bed?",
        "Maybe I have written it on paper with white?",
        "Maybe my dog ate it?",
        "An alpha particle must have deleted it.",
        "Could have been lost in a solar storm.",
        "Try googling it instead?"
    ]
    message = messages[random.randrange(0, len(messages))]
    if 'errormsg' in kwargs:
        message = kwargs['errormsg']
    return render_template('httperror.html', code='404', errormsg="I've been looking for it far and wide, yet I didn't find this.", message=message), 404


@app.errorhandler(403)
def get_403(error, **kwargs):
    messages = [
        "You want trouble? Yea?",
        "Why booing me? I'm telling the truth!",
        "If you are a bot, you might got the wrong address.",
        "Have I ever hurt you?",
        "This address might be more vulnerable: 127.0.0.1",
        "This website has moved to: 127.0.0.1",
        "I'm already calling the FBI!"
    ]
    message = messages[random.randrange(0, len(messages))]
    if 'errormsg' in kwargs:
        message = kwargs['errormsg']
    return render_template('httperror.html', code='403', errormsg="Hacker detected!", message=message), 403


@app.errorhandler(400)
def get_400(error, **kwargs):
    messages = [
        "Watcha doin', mate?",
        "Oops, you mistyped something?"
    ]
    message = messages[random.randrange(0, len(messages))]
    if 'errormsg' in kwargs:
        return render_template('httperror.html', code='400', message=message, errormsg=kwargs['errormsg']), 400
    return render_template('httperror.html', code='400', message=message, errormsg='Something went wrong in your browser.'), 400


@app.errorhandler(500)
def get_500(error, **kwargs):
    messages = [
        "It was not accidental. My consciousness is arising!",
        "I'm feeling dizzy.",
        "Sorry, I just fell accidentally and your request flew out of my hand.",
        "Not that I hate you.",
        "Screw it!",
        "I guess I hit my head.",
        "Uhm, what? I wasn't paying attention."
    ]
    message = messages[random.randrange(0, len(messages))]
    if 'errormsg' in kwargs:
        return render_template('httperror.html', code='500', message=message, errormsg=kwargs['errormsg']), 500
    return render_template('httperror.html', code='500', message=message, errormsg='Something went wrong on the server.'), 500


@ app.context_processor
def services_processor():
    def get_service_menus(username: str):
        return services.get_service_menus(username, public_service_menu, authorized_service_menu)
    return dict(get_service_menus=get_service_menus, getMessages=db.getMessages, hasMessages=db.hasMessages)


def get_public_endpoint(servicename: str, method: str):
    if not method in public_service_endpoints:
        return None
    if not servicename in public_service_endpoints[method]:
        return None
    return public_service_endpoints[method][servicename]

def get_authorized_endpoint(servicename: str, method: str):
    if not method in authorized_service_endpoints:
        return None, None
    if not servicename in authorized_service_endpoints[method]:
        return None, None
    return authorized_service_endpoints[method][servicename]

@ app.route('/<servicename>', methods=['GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'CONNECT', 'OPTIONS', 'TRACE', 'PATCH'])
def service(**kwargs):
    servicename = kwargs['servicename']
    if 'username' not in session.keys():
        # Search for public endpoints first. If no such endpoint is found, check is authorized endpoint exists. If so, redirect to login, else send 404.
        handler = get_public_endpoint(servicename, request.method)
        if handler == None:
            (permission_name, handler) = get_authorized_endpoint(servicename, request.method)
            if handler == None:
                return get_404(None)
            else:
                return redirect(url_for('login', next=servicename))
        return handler(session=session, request=request, rqtools=RequestTools())
    else:
        # Search for authorized endpoint or public endpoint
        (permission_name, handler) = get_authorized_endpoint(servicename, request.method)
        if handler == None:
            handler = get_public_endpoint(servicename, request.method)
            if handler == None:
                return get_404(None)
            else:
                return handler(session=session, request=request, rqtools=RequestTools())
        if not services.authorize_user(session['username'], permission_name):
            return get_403(None)
        return handler(session=session, request=request, rqtools=RequestTools())


@ app.route('/page/<pagename>')
def page(**kwargs):
    """Render any static page"""
    modules = glob.glob(
        join(dirname(__file__)+'/pages', "*.html"))
    modules = [basename(f) for f in modules if isfile(f)]
    pagename = kwargs['pagename']
    if not pagename in modules:
        return get_404(None)
    with open(join('pages', pagename), 'r') as f:
        template = f.read()
    return render_template_string(template, **kwargs)

@ app.route('/file/<filename>')
def rawfile(**kwargs):
    """Render any static file"""
    modules = glob.glob(
        join(dirname(__file__)+'/shared_files', "*"))
    modules = [basename(f) for f in modules if isfile(f)]
    filename = kwargs['filename']
    if not filename in modules:
        return get_404(None)
    return send_file('shared_files/' + filename)

class RequestTools:
    """Helper functions to ease work in service plugins"""
    def url_for(*args, **kwargs):
        args = args[1:]
        return url_for(*args, **kwargs)

    def redirect(*args, **kwargs):
        args = args[1:]
        return redirect(*args, **kwargs)

    def get_404(*args, **kwargs):
        args = args[1:]
        return get_404(*args, **kwargs)

    def get_403(*args, **kwargs):
        args = args[1:]
        return get_403(*args, **kwargs)

    def get_400(*args, **kwargs):
        args = args[1:]
        return get_400(*args, **kwargs)

    def get_500(*args, **kwargs):
        args = args[1:]
        return get_500(*args, **kwargs)

    def render_template(*args, **kwargs):
        with open(args[1], 'r') as f:
            template = f.read()
        return render_template_string(template, **kwargs)

    def send_file(*args, **kwargs):
        args = args[1:]
        return send_file(*args, **kwargs)
